# Class to record all desired redis activity

import base64
import fnmatch
import time
import tqdm
import redis

from .. import util
from ..storage.recorder import get_recorder
from ..config import *
from .. import cmd


def record(
        name=None, stream_ids='*', ignore_streams=[], 
        record_key=RECORD_KEY, 
        pause_key=RECORD_PAUSE_KEY,
        out_dir=RECORDING_DIR, 
        stream_refresh=3000, data_block=1000, wait_block=3000, no_streams_sleep=1000,
        single_recording=None, recording_type=STORAGE_FORMAT,
        host=HOST, port=PORT, db=DB
):
    '''Record redis streams to file.
    
    Arguments:
        name (str): name of the recording. Only use this if you want to start a recording when the script begins.
        stream_ids (list[str]): Stream IDs or Stream ID patterns (``["prefix:*"]``) to record. 
        ignore_streams (list[str]): Stream IDs or Stream ID patterns (``["prefix:*"]``) to ignore. 
        record_key (str): the key in redis used to store the recording name.
        out_dir (str): the directory to write recordings to.
        stream_refresh (int): the number of milliseconds between querying for new streams.
        data_block (int): How long to block for data (in milliseconds).
        wait_block (int): How long to block for recording name changes (in milliseconds). It only blocks if no recording is currently happening.
        no_streams_sleep (int): How long to sleep (in milliseconds) if no streams are available. 
    '''
    record_name = name  # 
    single_recording = bool(name) if single_recording is None else single_recording

    # initialize stream IDs
    stream_ids = (stream_ids or '*').split('+')
    stream_patterns = [s for s in stream_ids if '*' in s]
    stream_ids = [s for s in stream_ids if '*' not in s]
    tqdm.tqdm.write(f"Streams: {stream_ids}. Stream Patterns: {stream_patterns}")
    assert stream_ids or stream_patterns

    if ignore_streams is not False:
        ignore_streams = list(ignore_streams) + list(IGNORE_STREAMS)

    r = redis.Redis(host=host, port=port, db=db)
    pbar = tqdm.tqdm()
    paused = False
    try:
        with get_recorder(type=recording_type, out_dir=out_dir) as rec, r:
            # initialize cursor
            cursor = {s: '$' for s in stream_ids}
            rec_cursor = {record_key: '0', pause_key: '0'}
            
            # initialize recording state
            record_start_timestamp = None
            if record_name:
                record_start_timestamp = util.format_epoch_time(time.time())
                rec_cursor[record_key] = record_start_timestamp
                tqdm.tqdm.write(f"Starting recording: {record_name}")
                cmd.start_recording(r, record_name, record_start_timestamp)

            t0s = 0
            while True:
                # ---------------------- Watch for changes in recording ---------------------- #

                # query for recording name stream
                results, rec_cursor = util.read_latest(r, rec_cursor, block=False if record_name else wait_block)

                # check for recording changes
                for sid, xs in results:
                    # check for pause command
                    if sid == pause_key:
                        for t, x in xs[-1:]:
                            paused = int(x[b'd'].decode() if x else None or 0)
                            for c in cursor:
                                cursor[c] = t
                    # check for recording name command
                    if sid == record_key:
                        for t, d in xs[-1:]:
                            # get new recording name
                            x = (d.get(b'd') or b'').decode() or None
                            paused = int(d.get(b'paused', b'').decode() or 0)

                            if record_name != x:  # idempotent

                                # ------------------------------- Finishing up ------------------------------- #

                                if record_name:  # finish up active recording
                                    rec.ensure_writer(record_name)
                                    tstamp = util.parse_epoch_time(t)

                                    while cursor:
                                        # read data from redis and write to file
                                        results, cursor = util.read_next(r, cursor, block=data_block)
                                        result_sids = {s for s, x in results if x}
                                        for sid, xs in results:
                                            for t, x in xs:
                                                # the data is from after the recording ended.
                                                if util.parse_epoch_time(t) > tstamp:
                                                    result_sids.remove(sid)
                                                    break
                                                # write data
                                                pbar.set_description(f'{sid} {t}')
                                                rec.write(sid, util.parse_epoch_time(t), x)
                                                pbar.update()

                                        # update cursor to only include active streams
                                        cursor = {k: v for k, v in cursor.items() if k in result_sids}

                                rec.close()

                                # ----------------------------- Change recording ----------------------------- #

                                record_name = x
                                record_start_timestamp = t
                                cursor = {sid: record_start_timestamp for sid in stream_ids}

                                tqdm.tqdm.write(f'{"new recording:" if record_name else "ended recording"}: {record_name} {record_start_timestamp}')

                # no recording...
                if not record_name or paused:
                    continue

                # --------------------------- Query for new streams -------------------------- #

                # keep up-to-date list of streams
                t1s = time.time()
                if stream_patterns and t1s - t0s > stream_refresh / 1000:
                    keys = {x.decode('utf-8') for x in r.scan_iter(match=stream_ids or None, _type='stream')}
                    for k in keys - set(cursor):
                        if (
                            k != record_key and
                            not any(fnmatch.fnmatch(k, p) for p in ignore_streams) and
                            any(fnmatch.fnmatch(k, p) for p in stream_patterns)
                        ):
                            tqdm.tqdm.write(f"adding stream: {k} {record_start_timestamp}")
                            cursor[k] = record_start_timestamp
                    t0s = t1s

                # no streams to record, just wait.
                if not cursor:
                    # we sleep but we have the record start timestamp so we won't lose any data.
                    # We just can't sleep too long and fall behind
                    tqdm.tqdm.write("no cursor. sleeping")
                    time.sleep(no_streams_sleep/1000)
                    continue

                # ---------------------------- Pull data and write --------------------------- #

                # read data from redis and write to file
                results, cursor = util.read_next(r, cursor, block=data_block)
                rec.ensure_writer(record_name)
                for sid, xs in results:
                    for t, x in xs:
                        pbar.set_description(f'{sid} {t}')
                        rec.write(sid, util.parse_epoch_time(t), data=x)
                        pbar.update()
    finally:
        if single_recording:
            cmd.stop_recording(r)



def cli():
    import logging
    from tqdm.contrib.logging import logging_redirect_tqdm
    import fire
    logging.basicConfig(level=logging.DEBUG)
    with logging_redirect_tqdm():
        fire.Fire(record)

if __name__ == '__main__':
    from pyinstrument import Profiler
    p = Profiler()
    try:
        with p:
            cli()
    finally:
        p.print()



# Class to record all desired redis activity
import os
import base64
import re
import tqdm
import codecs
import redis
from redis.client import Monitor as Monitor_

from .. import util
from ..storage.recorder import get_recorder
from ..config import *
from .. import cmd




class Monitor(Monitor_):
    monitor_re = re.compile(rb"\[(\d+) (.*?)\] (.*)")
    command_re = re.compile(rb'"(.*?(?<!\\))"')
    def next_command(self):
        """Parse the response from a monitor command"""
        # https://github.com/redis/redis/blob/4031a187321be26fc96c044fec3ee759841e8379/src/replication.c#L576
        response = self.connection.read_response(disable_decoding=True)
        command_time, command_data = response.split(b" ", 1)
        db_id, client_info, raw_command = self.monitor_re.match(command_data).groups()
        command_parts = self.command_re.findall(raw_command)  # TODO: this is slow
        command_parts = [codecs.escape_decode(x)[0] for x in command_parts]

        client_info = client_info.decode()
        if client_info == "lua":
            client_address = client_type = "lua"
            client_port = ""
        else:
            client_address, client_port = client_info.rsplit(":", 1)  # ipv6 has colons
            client_type = "unix" if client_info.startswith("unix") else "tcp"
        return {
            "time": float(command_time.decode()),
            "db": int(db_id.decode()),
            "client_address": client_address,
            "client_port": client_port,
            "client_type": client_type,
            "args": command_parts,
        }



def record(
        name=None, record_key=MONITOR_KEY, record_cmds=MONITOR_CMDS, 
        out_dir=RECORDING_DIR, single_recording=None, recording_type='mcap',
        host=HOST, port=PORT, db=DB,
    ):
    '''Record all commands using MONITOR command.
    
    Arguments:
        name (str): name of the recording. Only use this if you want to start a recording when the script begins.
        record_key (str): the key in redis used to store the recording name.
        record_cmds (list[str]): the redis commands to record.
        out_dir (str): the directory to write recordings to.
    '''
    record_name = name
    single_recording = bool(name) if single_recording is None else single_recording

    pbar = tqdm.tqdm()

    record_cmds = [c.lower() for c in record_cmds]
    tqdm.tqdm.write(f"Monitoring commands: {record_cmds}")

    r = redis.Redis(host=host, port=port, db=db, decode_responses=False)
    try:
        with get_recorder(type=recording_type, schema={ "cmd": { "type": "array" } }, out_dir=out_dir) as rec, r:
            if record_name:
                cmd.start_monitoring(r, record_name)

            with Monitor(r.connection_pool) as m:
                # look up the initial recording name
                record_name = (r.get(record_key) or b'').decode()
                tqdm.tqdm.write(f"recording name: {record_name}")

                for data in m.listen():
                    # parse command
                    cmd_items = data['args']
                    if not cmd_items:  # rare parsing problems
                        continue
                    cmd_name = cmd_items[0].decode().lower()
                    if cmd_name not in CMD_IGNORE_DISPLAY:
                        pbar.set_description(cmd_name)

                    # listen for delete key
                    if cmd_name == 'del':
                        for key in cmd_items[1:]:
                            if key == record_key:
                                rec.close()
                                record_name = None
                                print("ended recording")
                                continue

                    # listen for set key
                    if cmd_name == 'set':
                        key, value = cmd_items[1:3]
                        if key == record_key and record_name != value:
                            rec.close()
                            record_name = value
                            print("new recording:" if record_name else "ended recording", record_name)
                            continue
                    
                    # no recording
                    if not record_name:
                        continue
                    
                    # make sure a writer is open
                    rec.ensure_writer(record_name)
                    if cmd_name not in record_cmds:
                        continue

                    # pick which channel to write to
                    if cmd_name in {'set', 'xadd'}:
                        sid = b'::'.join(cmd_items[:2]).decode()
                    else:
                        sid = cmd_items[0].decode()

                    # write command to file
                    rec.write(sid, data['time'], prepare_data(cmd_items))
                    pbar.update()
    finally:
        if single_recording:
            cmd.stop_monitoring(r)


def prepare_data(cmd_items):
    return {"cmd": [base64.b64encode(x).decode() for x in cmd_items]}



def cmd():
    import fire
    fire.Fire(record)

if __name__ == '__main__':
    from pyinstrument import Profiler
    p = Profiler()
    try:
        with p:
            cmd()
    finally:
        p.print()



import os
import zipfile
from .base import BaseRecorder
from ...util import format_epoch_time, move_with_suffix

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


MB = 1024 * 1024

class ZipRecorder(BaseRecorder):
    def __init__(self, out_dir='.', max_len=1000, max_size=9.5*MB):
        super().__init__()
        self.out_dir = out_dir
        self.max_len = max_len
        self.max_size = max_size
        self.recording_dir = None
        self.writer = None

    def write(self, stream_id, timestamp, data):
        assert set(data) == {b'd'}, f"zip recorder can only record a single field in a stream. got {set(data)}"
        self.ensure_channel(stream_id)
        timestamp = self._fix_timestamp(timestamp)
        self.writer[stream_id].write(data[b'd'], timestamp)

    def close(self):
        if self.writer:
            for _, w in self.writer.items():
                w.close()
            log.info("Closed Recorder: %s", self.recording_dir)
        self.writer = None

    def ensure_writer(self, record_name, force=False):
        if self.writer is not None and force:
            self.close()
        if self.writer is None:
            self.writer = {}
            self.recording_dir = os.path.join(self.out_dir, record_name)
            os.makedirs(self.recording_dir, exist_ok=True)
            log.info("Created Recorder: %s", self.recording_dir)
        return self

    def ensure_channel(self, channel):
        if channel not in self.writer:
            out_dir = os.path.join(self.recording_dir, channel)
            move_with_suffix(out_dir, prefix='_')
            self.writer[channel] = ZipWriter(out_dir, self.max_len, self.max_size)


MB = 1024 * 1024

class ZipWriter:
    def __init__(self, out_dir, max_len=1000, max_size=9.5*MB):
        super().__init__()
        self.buffer = []
        self.out_dir = out_dir
        self.max_len = max_len
        self.max_size = max_size
        self.size = 0

    def __enter__(self):
        pass

    def __exit__(self, e, t, tb):
        self.close()

    def write(self, data, ts):
        self.size += len(data)
        if not isinstance(ts, str):
            ts = format_epoch_time(ts)
        self.buffer.append([data, ts])
        if len(self.buffer) >= self.max_len or self.size >= self.max_size:
            self._dump(self.buffer)
            self.size = 0

    def _dump(self, data):
        if not data:
            return
        os.makedirs(self.out_dir, exist_ok=True)
        fname = os.path.join(self.out_dir, f'{data[0][1]}_{data[-1][1]}.zip')
        with zipfile.ZipFile(fname, 'a', zipfile.ZIP_STORED, False) as zf:
            for d, ts in data:
                zf.writestr(ts, d)
        data.clear()

    def close(self):
        if self.buffer:
            self._dump(self.buffer)



from redis_record.config import STORAGE_FORMAT


def get_recorder(*a, type=STORAGE_FORMAT, **kw):
    # XXX: not extensible
    if type == 'mcap':
        from .mcap import MCAPRecorder
        return MCAPRecorder(*a, **kw)
    if type == 'zip':
        from .zip import ZipRecorder
        return ZipRecorder(*a, **kw)
    
    raise ValueError(f"Unknown recorder type: {type}")


import os
import datetime
from ...config import *
from ...util import get_recording_filename


class MCAPInfo:
    def list(self, *, info=False, recording_dir=RECORDING_DIR):
        recordings = [
            os.path.splitext(f)[0]
            for f in os.listdir(recording_dir)
        ]
        if info:
            recordings = [self.info(n) for n in recordings]
        return recordings
    
    def info(self, name, recording_dir=RECORDING_DIR):
        from mcap.reader import make_reader
        fname = get_recording_filename(name, recording_dir)
        with open(fname, "rb") as f:
            reader = make_reader(f)
            s = reader.get_summary()
            stat = s.statistics
            
            t0 = stat.message_start_time / 10e8 # FIXME: ????????
            t1 = stat.message_end_time / 10e8
            return {
                "name": name,
                "start_time": t0,
                "end_time": t1,
                "start_datetime": datetime.datetime.fromtimestamp(t0).strftime('%c'),
                "end_datetime": datetime.datetime.fromtimestamp(t1).strftime('%c'),
                "streams": [
                    {
                        "name": c.topic,
                        "count": stat.channel_message_counts[c.id]
                    }
                    for c in s.channels.values()
                ],
                # "n_chunks": len(s.chunk_indexes),
                # "chunks": [c.chunk_length for c in s.chunk_indexes],
            }



import tqdm
import queue
import redis
import logging

from ..config import *
from ..util import read_latest
from ..sync import Sync
from ..storage.replay import get_player

log = logging.getLogger(__name__)

def replay_once(name, host=HOST, port=PORT, db=DB, realtime=True, speed_fudge=1, recording_type=STORAGE_FORMAT, recording_dir=RECORDING_DIR):
    r = redis.Redis(host=host, port=port, db=db)
    sync = Sync(speed_fudge=speed_fudge)

    pbar = tqdm.tqdm()
    with get_player(name, recording_dir, type=recording_type) as reader:
        for stream_id, ts, data in reader.iter_messages():
            if realtime:
                sync.sync(ts)

            pbar.update()
            pbar.set_description(f'{ts:.3f} {stream_id}')
            r.xadd(stream_id, data, '*') # t1

def replay(
        name=None,
        replay_key=REPLAY_KEY, 
        seek_key=REPLAY_SEEK_KEY,
        pause_key=REPLAY_PAUSE_KEY,
        recording_dir=RECORDING_DIR, 
        wait_block=3000,
        realtime=True, speed_fudge=1,
        single_recording=None, recording_type=STORAGE_FORMAT,
        host=HOST, port=PORT, db=DB
):
    if single_recording is None:
        single_recording = bool(name)
    if single_recording:
        return replay_once(name, host=host, port=port, db=db, realtime=realtime, speed_fudge=speed_fudge, recording_type=recording_type, recording_dir=recording_dir)

    r = redis.Redis(host=host, port=port, db=db)
    sync = Sync(speed_fudge=speed_fudge)
    pbar = tqdm.tqdm()

    player = None
    record_name = name
    paused = False
    seek = 0
    rec_cursor = {replay_key: '0'}
    while True:
        # ---------------------- Watch for changes in recording ---------------------- #

        # query for recording name stream
        results, rec_cursor = read_latest(r, rec_cursor, block=False if record_name else wait_block)

        # check for recording changes
        for sid, xs in results:
            if sid == replay_key:
                for t, x in xs[-1:]:
                    if player is not None:
                        player.close()
                        player = None
                    record_name = x[b'd'].decode() or None if x else None
                    paused = int(x.get(b'pause', b'0').decode() or 0)
                    seek = int(x.get(b'seek', b'0').decode() or 0)
                    if record_name:
                        log.info(f"new replay request: {record_name}")
            if sid == seek_key and player is not None:
                player.seek(int(x[b'd'].decode() if x else None or 0))
            if sid == pause_key:
                paused = int(x[b'd'].decode() if x else None or 0)

        if not record_name or paused:        
            continue

        if player is None:
            log.info(f'new ({recording_type}) player: {record_name} from {recording_dir}')
            try:
                player = get_player(record_name, recording_dir, type=recording_type)
            except FileNotFoundError as e:
                log.warning(f"File not Found: {e}")
                r.xadd(replay_key, {b'd': b''}, '*')
                continue
            if seek:
                player.seek(seek)
        

        msg = player.next_message()            
        if msg is None:
            log.info(f"Finished replaying recording: {record_name}")
            record_name = None
            r.xadd(replay_key, {b'd': b''}, '*')
            continue
        stream_id, ts, data = msg

        if realtime:
            sync.sync(ts)

        pbar.update()
        pbar.set_description(f'{ts:.3f} {stream_id}')
        r.xadd(stream_id, data, '*') # t1


def cli():
    import logging
    from tqdm.contrib.logging import logging_redirect_tqdm
    import fire
    logging.basicConfig(level=logging.DEBUG)
    with logging_redirect_tqdm():
        fire.Fire(replay)

if __name__ == '__main__':
    cli()


from __future__ import annotations
import os
import datetime
from .config import *

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def maybe_encode(s: str|bytes, encoding='utf-8'):
    '''If the input is str, encode as utf-8 (or other).'''
    return s.encode(encoding) if isinstance(s, str) else s

def maybe_decode(s: str|bytes, encoding='utf-8'):
    '''If the input is bytes, decode as utf-8 (or other).'''
    return s.decode(encoding) if isinstance(s, bytes) else s

# ---------------------------------------------------------------------------- #
#                               Timestamp parsing                              #
# ---------------------------------------------------------------------------- #


def parse_epoch_time(tid: str|bytes):
    '''Convert a redis timestamp to epoch seconds.'''
    return int(maybe_decode(tid).split('-')[0])/1000

def format_epoch_time(tid: float, i='0'):
    '''Format a redis timestamp from epoch seconds.'''
    return f'{int(tid * 1000)}-{i}'

def parse_datetime(tid: str|bytes):
    '''Convert a redis timestamp to a datetime object.'''
    return datetime.datetime.fromtimestamp(parse_epoch_time(tid))

def format_datetime(dt: datetime.datetime):
    '''Format a redis timestamp from a datetime object.'''
    return format_epoch_time(dt.timestamp())

def format_iso(tid: str|bytes, format: str=''):
    '''Convert a redis timestamp to a iso format.'''
    if not format:
        return parse_datetime(tid).strftime(format)
    return parse_datetime(tid).isoformat()

def nonspecific_timestamp(tid: str|bytes):
    return f"{maybe_decode(tid).split('-')[0]}-*"

# ---------------------------------------------------------------------------- #
#                                Data formatting                               #
# ---------------------------------------------------------------------------- #

def pack_entries(entries):
    offsets = []
    content = bytearray()
    for sid, data in entries:
        sid = maybe_decode(sid)
        for ts, d in data:
            content += d[b'd']
            offsets.append((sid, maybe_decode(ts), len(content)))
    # jsonOffsets = orjson.dumps(offsets).decode('utf-8')
    return offsets, content

def decode_xread_format(data):
    return [
        (maybe_decode(s), [(maybe_decode(t), x) for t, x in xs])
        for s, xs in data
    ]


# ---------------------------------------------------------------------------- #
#                                 Redis Reading                                #
# ---------------------------------------------------------------------------- #


def read_next(r, sids, block=0, count=1, **kw):
    data = r.xread(sids, block=block, count=count, **kw)
    data = decode_xread_format(data)
    sids = update_cursor(sids, data)
    return data, sids

def read_latest(r, sids, block=False, count=1, **kw):
    with r.pipeline() as p:
        for sid, t in sids.items():
            p.xrevrange(sid, '+', f'({t}', count=count, **kw)
        data = list(zip(sids, p.execute()))
    if block and not any(x for s, x in data):
        data = r.xread(sids, block=block, count=count, **kw)
    data = decode_xread_format(data)
    sids = update_cursor(sids, data)
    return data, sids


def update_cursor(sids: dict[str, str], data: list[str|tuple]) -> dict[str, str]:
    for s, ts in data:
        if ts:
            sids[s] = max(t for t, x in ts)
    return sids


# --------------------------------- Recording -------------------------------- #


def get_recording_filename(name, recording_dir=RECORDING_DIR):
    if os.path.isfile(name):
        return name
    return os.path.join(recording_dir, f'{name}.mcap')


def move_with_suffix(src, prefix=None, suffix=None, has_ext=None):
    if os.path.exists(src):
        has_ext = os.path.isfile(src) if has_ext is None else has_ext
        dir = os.path.dirname(src)
        base = os.path.basename(src)
        if has_ext:
            base, ext = os.path.splitext(base)
        else:
            base, ext = base, ''
        base = f"{dir}/{prefix or ''}{base}{suffix or ''}"

        i = 1
        dest = f"{base}_{i}{ext}"
        while os.path.exists(dest):
            i += 1
            dest = f"{base}_{i}{ext}"
        log.info("MOVING %s to %s", src, dest)
        os.rename(src, dest)



import os
import orjson
from .base import BaseRecorder
from ...util import move_with_suffix

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class JsonRecorder(BaseRecorder):
    recording_dir = ''
    def __init__(self, out_dir='.', list_key='data'):
        self.out_dir = out_dir
        self.writer = {}
        self.list_key = list_key

    def __enter__(self):
        return self

    def __exit__(self, *a):
        self.close()

    def ensure_writer(self, name):
        self.close()
        self.recording_dir = os.path.join(self.out_dir, name)
        move_with_suffix(self.recording_dir)
        os.makedirs(self.recording_dir, exist_ok=True)

    def ensure_channel(self, sid):
        if sid not in self.writer:
            self.writer[sid] = open(os.path.join(self.recording_dir, f'{sid}.json'), 'wb')
            self.writer[sid].write(b'[\n')
            self.writer[sid].i_line = 0
        return self.writer[sid]

    def close(self):
        for w in self.writer.values():
            w.write(b'\n]\n')
            w.close()
        self.writer.clear()

    def write(self, sid, t, data):
        w = self.ensure_channel(sid)
        data = self._add_timestamp_to_json(data, t)
        data = self._serialize(data)
        if w.i_line:
            w.write(b',\n')
        w.write(data)
        w.i_line += 1

    def _add_timestamp_to_json(self, d, ts):
        if ts is not None:
            try:
                if isinstance(d, bytes):
                    d = orjson.loads(d)
                if self.list_key:
                    if not isinstance(d, dict):
                        d = {self.list_key: d}
                    if 'timestamp' not in d:
                        d['timestamp'] = ts
            except Exception:
                print("error reading data")
                print(d)
                raise
        return d

    def _serialize(self, d):
        try:
            if not isinstance(d, bytes):
                d = orjson.dumps(d, option=orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY)
        except Exception:
            print("error writing data")
            print(d)
            raise
        return d



import time
import tqdm
import datetime
import redis
from .config import *
from .util import get_recording_filename


def _xadd(r, key, value, timestamp='*', maxlen=RECORD_NAME_MAXLEN, **kw):
    return r.xadd(key, {RECORD_XADD_DATA_FIELD: value, **kw}, id=timestamp or '*', maxlen=maxlen)

# ------------------------------ Monitor record ------------------------------ #

def start_monitoring(r, name):
    return r.set(MONITOR_KEY, name)

def stop_monitor(r):
    return r.delete(MONITOR_KEY)

# ---------------------------------- Record ---------------------------------- #

def start_recording(r, name, timestamp=None):
    return _xadd(r, RECORD_KEY, name, timestamp)

def stop_recording(r, timestamp=None):
    return _xadd(r, RECORD_KEY, '', timestamp)

def pause_recording(r, timestamp=None):
    return _xadd(r, RECORD_PAUSE_KEY, '1', timestamp)

def resume_recording(r, timestamp=None):
    return _xadd(r, RECORD_PAUSE_KEY, '0', timestamp)

# ---------------------------------- Replay ---------------------------------- #

def start_replay(r, name, timestamp=None, seek=0, pause=0):
    return _xadd(r, REPLAY_KEY, name, timestamp, seek=seek, pause=pause)

def stop_replay(r, timestamp=None):
    return _xadd(r, REPLAY_KEY, '', timestamp)

def pause_replay(r, timestamp=None):
    return _xadd(r, REPLAY_PAUSE_KEY, '1', timestamp)

def resume_replay(r, timestamp=None):
    return _xadd(r, REPLAY_PAUSE_KEY, '0', timestamp)

def seek_replay(r, seek, timestamp=None):
    return _xadd(r, REPLAY_SEEK_KEY, seek, timestamp)

def seek_replay(r, seek, timestamp=None):
    return _xadd(r, REPLAY_SEEK_KEY, seek, timestamp)



class Commands:
    def __init__(self, host=HOST, port=PORT, db=DB, r=None):
        self.r = r or redis.Redis(host, port=port, db=db)

    def __enter__(self):
        return self
    
    def __exit__(self, *a):
        self.r.close()

    # --------------------------------- Recording -------------------------------- #

    def start(self, name=None, timestamp=None):
        if not name:
            name = datetime.datetime.now().strftime("%Y.%m.%d-%H.%M.%S")
        return start_recording(self.r, name, timestamp)
    
    def pause_recording(self, timestamp=None):
        return pause_recording(self.r, timestamp)

    def resume_recording(self, timestamp=None):
        return resume_recording(self.r, timestamp)
    
    def stop(self, timestamp=None):
        return stop_recording(self.r, timestamp)
    
    # ---------------------------------- Replay ---------------------------------- #
    
    def replay(self, name, timestamp=None):
        return start_replay(self.r, name, timestamp)
    
    def pause(self, timestamp=None):
        return pause_replay(self.r, timestamp)
    
    def resume(self, timestamp=None):
        return resume_replay(self.r, timestamp)
    
    def seek(self, timestamp=None):
        return resume_replay(self.r, timestamp)
    
    def stop_replay(self, timestamp=None):
        return stop_replay(self.r, timestamp)
    
    def current_replay(self):
        return self.r.xrevrange(REPLAY_KEY, count=1)
    
    

    def recorder(self, *a, **kw):
        from redis_record.record import record
        return record(*a, **kw)
    
    def recorder(self, *a, **kw):
        from redis_record.replay import replay
        return replay(*a, **kw)
    
    # ----------------------------------- Misc ----------------------------------- #

    # def fake(self, *a, **kw):
    #     from .fake import data
    #     return data(*a, **kw)
    
    def fake(self, stream_id, rate=100, limit=None, **kw):
        from .sync import Clock
        sync = Clock(rate)
        pbar = tqdm.tqdm()
        i = 0
        while True:
            self.add_fake(stream_id, i, **kw)
            pbar.update()
            sync.sync()
            i += 1
            if limit and i >= limit:
                break
    
    def add_fake(self, stream_id, data=None, size=1000, **kw):
        x = str(data).encode() if data is not None else b'0'*size
        return x, _xadd(self.r, stream_id, x, **kw)

    def read(self, sids, block=None, **kw) -> tuple[list, dict[str, str]]:
        if isinstance(sids, list):
            sids = {s: f'{time.time()*1000:.0f}-0' for s in sids}
        data = self.r.xread(sids, block=block, **kw)
        # decode stream IDs and timestamps
        data = _decode_xread_format(data)
        return data, _update_cursor(sids, data)


def _update_cursor(sids: dict[str, str], data: list[str|tuple]) -> dict[str, str]:
    for s, ts in data:
        if ts:
            sids[s] = max(t for t, x in ts)
    return sids


def _decode_xread_format(data):
    return [(s.decode(), [(t.decode(), x) for t, x in xs]) for s, xs in data]


def cli():
    import fire
    fire.Fire(Commands)

if __name__ == '__main__':
    cli()



import os

_all_set_cmds = 'xadd+set+hmset+hset+hsetnx+lset+mset+msetnx+psetex+setbit+setrange+setex+setnx+getset+json.set+json.mset'

MONITOR_KEY = os.getenv("RECORD_MONITOR_KEY") or 'RECORD:NAME'
MONITOR_CMDS = (os.getenv("RECORD_MONITOR_CMDS") or _all_set_cmds).split("+")


RECORD_KEY = os.getenv("RECORD_STREAMS_KEY") or 'XRECORD:NAME'
REPLAY_KEY = os.getenv("REPLAY_STREAMS_KEY") or 'XREPLAY:NAME'
RECORD_PAUSE_KEY = os.getenv("RECORD_STREAMS_PAUSE_KEY") or 'XRECORD:PAUSE'
REPLAY_PAUSE_KEY = os.getenv("REPLAY_STREAMS_PAUSE_KEY") or 'XREPLAY:PAUSE'
REPLAY_SEEK_KEY = os.getenv("REPLAY_STREAMS_SEEK_KEY") or 'XREPLAY:SEEK'

RECORD_XADD_DATA_FIELD = (os.getenv("RECORD_STREAMS_DATA_KEY") or 'd').encode()
RECORD_NAME_MAXLEN = int(os.getenv("RECORD_NAME_MAXLEN") or 100)

HOST = os.getenv('REDIS_HOST') or 'localhost'
PORT = os.getenv('REDIS_PORT') or 6379
DB = os.getenv('REDIS_DB') or 0

RECORDING_DIR = os.getenv('RECORDING_DIR') or './recordings'
DEFAULT_CHANNEL = 'all'

IGNORE_STREAMS = {RECORD_KEY, REPLAY_KEY, RECORD_PAUSE_KEY, REPLAY_PAUSE_KEY, REPLAY_SEEK_KEY}
CMD_IGNORE_DISPLAY = {'multi', 'exec'}
STORAGE_FORMAT = os.getenv('RECORD_STORAGE_FORMAT') or 'zip'



import time
import datetime
import asyncio
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

class Sync:
    def __init__(self, speed_fudge=1, warn_above=10):
        self.speed_fudge = speed_fudge
        self.t0 = self.tproc = self.tpub0 = self.tpub = 0
        self.warn_above = warn_above

    def clear(self):
        self.t0 = self.tproc = self.tpub0 = self.tpub = 0

    def start(self, timestamp):
        self.t0 = self.tproc = time.time()
        self.tpub0 = self.tpub = timestamp

    def wait_time(self, timestamp):
        if self.t0 == 0:
            self.start(timestamp)

        ti = time.time()
        # estimate amount of time to sleep
        data_time_since_start = timestamp - self.tpub0
        real_time_since_start = ti - self.t0
        process_time = ti - self.tproc
        delay = max(data_time_since_start - real_time_since_start - process_time, 0)
        return delay

    def sync(self, timestamp):
        delay = self.wait_time(timestamp)
        if delay:
            self._warn(timestamp, delay)
            time.sleep(delay / self.speed_fudge)
        
        # update for next iteration
        self.tpub = timestamp
        self.tproc = time.time()

    async def sync_async(self, timestamp):
        delay = self.wait_time(timestamp)
        if delay:
            self._warn(timestamp, delay)
            await asyncio.sleep(delay / self.speed_fudge)
        
        # update for next iteration
        self.tpub = timestamp
        self.tproc = time.time()

    def _warn(self, timestamp, delay):
        if delay > self.warn_above:
            log.warning(
                f"Sleeping for {delay:.3f} seconds. \n"
                f"Simulating {datetime.datetime.fromtimestamp(self.tpub).strftime('%H:%M:%S.%f')} - {datetime.datetime.fromtimestamp(timestamp).strftime('%H:%M:%S.%f')}")


class Clock(Sync):
    def __init__(self, rate=None):
        self.delta = 1/rate if rate else 0
        super().__init__()

    def wait_time(self, timestamp=None):
        if not self.delta: return 0
        return super().wait_time(self.tpub + self.delta)
    
    def sync(self, timestamp=None):
        return super().sync(self.tpub + self.delta)
    
    async def sync_async(self, timestamp=None):
        return await super().sync_async(self.tpub + self.delta)
    
    def nowait_sync(self, timestamp=None):
        if not self.wait_time(timestamp):
            self.sync(timestamp)



if __name__ == '__main__':
    def test(rate=3):
        import tqdm
        pbar = tqdm.tqdm()
        sync = Clock(rate)
        while True:
            pbar.update()
            sync.sync()
    import fire
    fire.Fire(test)


import os
import json
import base64
from mcap.reader import make_reader
from redis_record.config import RECORDING_DIR

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class MCAPPlayer:
    def __init__(self, name, recording_dir=RECORDING_DIR, subset=None) -> None:
        self.path = name if os.path.isfile(name) else os.path.join(recording_dir, f'{name}.mcap')
        self.subset = subset
        log.info("Created Player: %s - streams: %s", self.path, self.subset or 'all')

    def __enter__(self):
        self.fh = open(self.path, "rb")
        self.reader = make_reader(self.fh)
        print(self.reader.get_header())
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()

    def close(self):
        self.fh.close()
        log.info("Closed Player: %s - streams: %s", self.path, self.subset or 'all')

    def __iter__(self):
        yield from self.iter_messages()

    it = None
    def iter_messages(self):
        if self.it is None:
            self.it = iter(self.reader.iter_messages(topics=self.subset))
        for schema, channel, message in self.it:
            yield self._parse_message(message)

    def next_message(self):
        if self.it is None:
            self.it = iter(self.reader.iter_messages())
        try:
            schema, channel, message = next(self.it)
        except StopIteration:
            return
        return self._parse_message(message)

    def _parse_message(self, message):
        args = json.loads(message.data)
        stream_id = args['stream_id']
        data = {k: base64.b64decode(x.encode()) for k, x in args['data'].items()}
        ts = message.publish_time*10e-9 * 10e-3 # for some reason it's -12 not -9??
        return stream_id, ts, data


import setuptools

setuptools.setup(
    name='redis_record',
    version='0.0.6',
    description='redis streams recordings',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    author='Bea Steers',
    author_email='bea.steers@gmail.com',
    url=f'https://github.com/beasteers/redis-record',
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['redis-record=redis_record.cmd:cli']},
    install_requires=[
        'redis', 'mcap', 'orjson', 'tqdm', 'fire',
    ],
    extras_require={},
    license='MIT License',
    keywords='redis record streams video streaming')



import os
import glob
import zipfile
import queue
from fnmatch import fnmatch

from redis_record.util import parse_epoch_time
from redis_record.config import RECORDING_DIR

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class ZipPlayer:
    def __init__(self, path, recording_dir=RECORDING_DIR, subset=None, raw_timestamp=None):
        self.recording_dir = path if recording_dir is None else os.path.join(recording_dir, path)
        self.subset = subset if isinstance(subset, (list, tuple, set)) else [subset] if subset else []
        self.file_index = {}
        self.file_cursor = {}
        self.time_cursor = 0
        self.zipfh = {}
        self.file_end_timestamps = {}
        self.queue = queue.PriorityQueue()
        self.raw_timestamp = raw_timestamp
        log.info("Created Player: %s - streams: %s", self.recording_dir, self.subset or 'all')

        self._load_file_index()
        for stream_id in self.file_index:
            self._queue_next_file(stream_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __iter__(self):
        yield from self.iter_messages()

    def seek(self, timestamp):
        self.close()  # TODO: do this more efficiently. the problem is selectively clearing the queue.
        for sid, fs in self.file_index.items():
            for i, (_, start, end) in enumerate(fs):
                if timestamp < end:
                    self._queue_file(sid, i)
                    break
        self.time_cursor = timestamp

    def next_message(self):
        # get next message
        while True:
            try:
                _, (stream_id, ts) = self.queue.get(block=False)
            except queue.Empty:
                return
            tx = parse_epoch_time(ts)
            if tx >= self.time_cursor or 0:
                # log.debug("using timestamp: %s", ts)
                break
            # log.debug("skipping timestamp: %s", ts)

        # load data
        self.time_cursor = tx
        with self.zipfh[stream_id].open(ts, 'r') as f:
            data = f.read()
        # possibly load next file
        if ts >= self.file_end_timestamps[stream_id]:
            self._queue_next_file(stream_id)
        tx = ts if self.raw_timestamp else tx
        return stream_id, tx, {'d': data}

    def iter_messages(self):
        msg = self.next_message()
        while msg:
            yield msg
            msg = self.next_message()

    def close(self):
        self.time_cursor = 0
        self.queue.queue.clear()
        for zf in self.zipfh.values():
            zf.close()
        self.zipfh.clear()
        log.info("Closed Player: %s - streams: %s", self.recording_dir, self.subset or 'all')

    def _get_time_range_from_file(self, fname):
        t0, t1 = fname.split(os.sep)[-1].removesuffix('.zip').split('_')
        return t0, t1

    def _load_file_index(self):
        self.file_index = {
            stream_id: [
                (f, *self._get_time_range_from_file(f))
                for f in sorted(glob.glob(os.path.join(self.recording_dir, stream_id, '*.zip')))
            ]
            for stream_id in os.listdir(self.recording_dir)
            if not self.subset or any(fnmatch(stream_id, p) for p in self.subset)
        }
        self.file_cursor = {s: -1 for s in self.file_index}
        log.debug('File Index: %s', self.file_index)
        log.debug('File Cursor: %s', self.file_cursor)

    def _queue_next_file(self, stream_id):
        self._queue_file(stream_id, self.file_cursor[stream_id] + 1)

    def _queue_file(self, stream_id, index):
        # close previous file
        if self.zipfh.get(stream_id) is not None:
            log.debug("closing opened file: %s", stream_id)
            self.zipfh.pop(stream_id).close()
        # check if we're done with this stream
        if index >= len(self.file_index[stream_id]):
            log.debug("finished stream: %s", stream_id)
            return
        
        # load the zipfile
        fname, _, _ = self.file_index[stream_id][index]
        log.debug("queueing file: %s", fname)
        self.file_cursor[stream_id] = index
        self.zipfh[stream_id] = zf = zipfile.ZipFile(fname, 'r', zipfile.ZIP_STORED, False)
        # load the file list
        ts = sorted(set(zf.namelist()))
        if not ts:
            log.debug("no data in file: %s", fname)
            self.zipfh[stream_id].close()
            return 
        self.file_end_timestamps[stream_id] = ts[-1]

        # queue up the timestamps
        log.debug("putting %d timestamps - up to %s", len(ts), ts[-1])
        for t in ts:
            tx = parse_epoch_time(t)
            self.queue.put((int(tx*1e6), (stream_id, t)))


def replay(name, *a, **kw):
    with ZipPlayer(name, *a, **kw) as player:
        for sid, t, data in player.iter_messages():
            logging.info('%s %s %s', sid, t, data)

def cli():
    import logging
    from tqdm.contrib.logging import logging_redirect_tqdm
    import fire
    logging.basicConfig(level=logging.DEBUG)
    with logging_redirect_tqdm():
        fire.Fire(replay)

if __name__ == '__main__':
    cli()



import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)



class BaseRecorder:
    def __init__(self, out_dir='.', schema=None):
        self.out_dir = out_dir
        self.schema = schema
        self.last_timestamp = None
        self.index = 0

    def _fix_timestamp(self, timestamp):
        if isinstance(timestamp, str):
            return timestamp

        timestamp = int(timestamp*1000)
        if self.last_timestamp == timestamp:
            self.index += 1
        else:
            self.index = 0
        self.last_timestamp = timestamp
        return f'{timestamp}-{self.index}'

    def __enter__(self):
        return self

    def __exit__(self, *a):
        self.close()

    def write(self, sid, timestamp, data):
        raise NotImplementedError

    def close(self):
        pass

    def ensure_writer(self, name, force=False):
        raise NotImplementedError

    def ensure_channel(self, sid):
        raise NotImplementedError



import os
import datetime
from ...config import *
from ...util import get_recording_filename


class ZIPInfo:
    def __init__(self, recording_dir=RECORDING_DIR):
        self.recording_dir = recording_dir

    def list(self, *, info=False):
        recordings = [
            os.path.splitext(f)[0]
            for f in os.listdir(self.recording_dir)
        ]
        if info:
            recordings = [self.info(n) for n in recordings]
        return recordings
    
    def info(self, name):
        path = os.path.join(self.recording_dir, name)
        sids = os.listdir(path)
        files = {sid: os.listdir(os.path.join(path, sid)) for sid in sids}
        times = {sid: [tuple(float(t) for t in f.removesuffix('.zip').split('_')) for f in fs] for sid, fs in files.items()}
        starts, ends = zip(*times) or ((), ())
        t0 = min(starts)
        t1 = max(ends)

        return {
            "name": name,
            "start_time": t0,
            "end_time": t1,
            "start_datetime": datetime.datetime.fromtimestamp(t0).strftime('%c'),
            "end_datetime": datetime.datetime.fromtimestamp(t1).strftime('%c'),
            "streams": [
                {
                    "name": sid,
                    "chunks": times,
                }
                for sid, fs in files.items()
            ],
        }



# Class to record all desired redis activity

import os
import time
import json
import base64
from mcap.writer import Writer
from .base import BaseRecorder
from ...config import DEFAULT_CHANNEL
from ...util import move_with_suffix

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class MCAPRecorder(BaseRecorder):
    def __init__(self, schema=None, out_dir='.'):
        self.out_dir = out_dir
        self.writer = self.fhandle = None
        self.schema = schema
        self.channel_ids = {}

    def __enter__(self):
        return self

    def __exit__(self, *a):
        self.close()

    def ensure_writer(self, record_name, force=False):
        if self.writer is not None and force:
            self.close()

        if self.writer is None:
            # create new writer
            os.makedirs(self.out_dir or '.', exist_ok=True)
            self.fname = fname = os.path.join(self.out_dir, f'{record_name}.mcap')
            move_with_suffix(self.fname)
            self.fhandle = fhandle = open(fname, "wb")
            self.writer = writer = Writer(fhandle)
            writer.start()
            log.info("Created Recorder: %s", fname)

            # https://mcap.dev/docs/python/mcap-apidoc/mcap.well_known#mcap.well_known.MessageEncoding
            kw = {"properties": self.schema} if self.schema else {}
            self.schema_id = writer.register_schema(
                name="data",
                encoding="jsonschema",
                data=json.dumps({"type": "object", **kw}).encode(),
            )
        return self

    def ensure_channel(self, channel=DEFAULT_CHANNEL):
        if channel not in self.channel_ids:
            log.debug("Opening Recorder Channel: %s %s", self.fname, channel)
            self.channel_ids[channel] = self.writer.register_channel(
                schema_id=self.schema_id,
                topic=channel,
                message_encoding="json",
            )

    def write(self, stream_id, timestamp, data):
        self.ensure_channel(stream_id)
        data = {'stream_id': stream_id, 'data': prepare_data(data)}
        self.writer.add_message(
            channel_id=self.channel_ids[stream_id],
            log_time=time.time_ns(),
            data=json.dumps(data).encode("utf-8"),
            publish_time=int(timestamp * 10e9),
        )

    def close(self):
        if self.writer is not None:
            self.writer.finish()
            self.fhandle.close()
            self.writer = self.fhandle = None
            self.channel_ids.clear()
            log.info("Closed Recorder: %s", self.fname)



def prepare_data(data):
    return {k.decode(): base64.b64encode(v).decode() for k, v in data.items()}



def read(fname):
    from mcap.reader import make_reader
    with open(fname, "rb") as f:
        reader = make_reader(f)
        print(reader.get_header())
        # summary = reader.get_summary()
        # print(summary)
        for schema, channel, message in reader.iter_messages():
            print(f"{channel.topic} ({schema.name}): {message.data}")


def cli():
    import fire
    fire.Fire({
        "read": read,
    })

if __name__ == '__main__':
    cli()


import redis
from redis_record.record.monitor import Monitor
from .config import *

def watch(host=HOST, port=PORT, db=DB, charlimit=30):
    with redis.Redis(host=host, port=port, db=db) as r:
        with Monitor(r.connection_pool) as m:
            for data in m.listen():
                cmd_items = data['args']
                print(*[x[:charlimit] for x in cmd_items], max(len(x) for x in cmd_items))


def cli():
    import fire
    fire.Fire(watch)

if __name__ == '__main__':
    cli()


import time
import json
import tqdm
import base64
import redis
from mcap.reader import make_reader

from ..config import *
from ..util import get_recording_filename


def replay(name, host=HOST, port=PORT, db=DB, realtime=True, speed_fudge=1, recording_dir=RECORDING_DIR):
    fname = get_recording_filename(name, recording_dir)

    r = redis.Redis(host=host, port=port, db=db)

    pbar = tqdm.tqdm()
    t0 = 0
    with open(fname, "rb") as f:
        reader = make_reader(f)
        print(reader.get_header())
        for schema, channel, message in reader.iter_messages():
            args = json.loads(message.data)['cmd']
            args = [base64.b64decode(x.encode()) for x in args]
            t1 = message.publish_time*10e-9 * 10e-3 # for some reason it's -12 not -9??

            # real-time
            if realtime:
                delay = max(t1 - t0, 0)
                if delay > 1:
                    pbar.set_description(f'sleeping for {delay}s {t1} {args[0]} - {t0}')
                if t0 and delay:
                    time.sleep(delay/speed_fudge) # FIXME recalc speed accounting for lost time. /speed_fudge is a quick fix.
                t0 = t1

            pbar.update()
            pbar.set_description(f'{t1:.3f} {b" ".join(args[:2])}')
            r.execute_command(*args)

def cli():
    import fire
    fire.Fire(replay)

if __name__ == '__main__':
    cli()


import redis
import tqdm
from .config import *
from .sync import Clock

def data(stream_id='fake_data', size=1000, rate=None, payload={}, maxlen=1000, limit=None, r=None, host=HOST, port=PORT, db=DB):
    sync = Clock(rate)
    with (r or redis.Redis(host=host, port=port, db=db)) as r:
        pbar = tqdm.tqdm()
        i = 0
        while True:
            x = b'0'*size
            r.xadd(stream_id, {RECORD_XADD_DATA_FIELD: x, **payload}, id='*', maxlen=maxlen)
            pbar.update()
            sync.sync()
            i += 1
            if limit and i > limit:
                break

def cli():
    import fire
    fire.Fire(data)

if __name__ == '__main__':
    cli()